@@ -21,7 +21,10 @@
 #include <atomic>
 #include <climits>
 #include <cstddef>
+#include <cstdint>
+#include <initializer_list>
 #include <memory>
+#include <optional>
 #include <string>
 #include <type_traits>
 #include <utility>
@@ -30,13 +33,16 @@
 #include "google/protobuf/stubs/common.h"
 #include "absl/base/call_once.h"
 #include "absl/base/casts.h"
+#include "absl/base/optimization.h"
 #include "absl/strings/string_view.h"
 #include "google/protobuf/any.h"
 #include "google/protobuf/has_bits.h"
 #include "google/protobuf/implicit_weak_message.h"
+#include "google/protobuf/internal_visibility.h"
 #include "google/protobuf/message_lite.h"
 #include "google/protobuf/port.h"
 #include "google/protobuf/repeated_field.h"
+#include "google/protobuf/repeated_ptr_field.h"
 #include "google/protobuf/wire_format_lite.h"
 
 
@@ -66,7 +72,7 @@ namespace internal {
 PROTOBUF_EXPORT extern std::atomic<bool> init_protobuf_defaults_state;
 PROTOBUF_EXPORT void InitProtobufDefaultsSlow();
 PROTOBUF_EXPORT inline void InitProtobufDefaults() {
-  if (PROTOBUF_PREDICT_FALSE(
+  if (ABSL_PREDICT_FALSE(
           !init_protobuf_defaults_state.load(std::memory_order_acquire))) {
     InitProtobufDefaultsSlow();
   }
@@ -321,6 +327,76 @@ struct WeakDescriptorDefaultTail {
   size_t size;
 };
 
+// Tag to distinguish overloads below:
+//  - if last argument is `BytesTag tag = BytesTag{}` then the overload is
+//    available to both string and byte fields.
+//  - if last argument is `BytesTag tag` then the overload is only available to
+//    byte fields.
+//  - if there is no BytesTag argument, then the overload is only available to
+//    string fields.
+struct BytesTag {
+  explicit BytesTag() = default;
+};
+
+// Assigns to `dest` the content of `value`, optionally bounded by `size`.
+// This overload set is used to implement `set_xxx()` methods for repeated
+// string fields in generated code.
+inline void AssignToString(std::string& dest, const std::string& value,
+                           BytesTag /*tag*/ = BytesTag{}) {
+  dest.assign(value);
+}
+inline void AssignToString(std::string& dest, std::string&& value,
+                           BytesTag /*tag*/ = BytesTag{}) {
+  dest.assign(std::move(value));
+}
+inline void AssignToString(std::string& dest, const char* value,
+                           BytesTag /*tag*/ = BytesTag{}) {
+  dest.assign(value);
+}
+inline void AssignToString(std::string& dest, const char* value,
+                           std::size_t size) {
+  dest.assign(value, size);
+}
+inline void AssignToString(std::string& dest, const void* value,
+                           std::size_t size, BytesTag /*tag*/) {
+  dest.assign(reinterpret_cast<const char*>(value), size);
+}
+inline void AssignToString(std::string& dest, absl::string_view value,
+                           BytesTag /*tag*/ = BytesTag{}) {
+  dest.assign(value.data(), value.size());
+}
+
+// Adds `value`, optionally bounded by `size`, as the last element of `dest`.
+// This overload set is used to implement `add_xxx()` methods for repeated
+// string fields in generated code.
+template <typename Arg, typename... Args>
+void AddToRepeatedPtrField(InternalVisibility visibility, google::protobuf::Arena* arena,
+                           google::protobuf::RepeatedPtrField<std::string>& dest,
+                           Arg&& value, Args... args) {
+  AssignToString(*dest.InternalAddWithArena(visibility, arena),
+                 std::forward<Arg>(value), args...);
+}
+inline void AddToRepeatedPtrField(InternalVisibility visibility,
+                                  google::protobuf::Arena* arena,
+                                  google::protobuf::RepeatedPtrField<std::string>& dest,
+                                  std::string&& value,
+                                  BytesTag /*tag*/ = BytesTag{}) {
+  dest.InternalAddWithArena(visibility, arena, std::move(value));
+}
+
+constexpr std::optional<uintptr_t> EncodePlacementArenaOffsets(
+    std::initializer_list<size_t> offsets) {
+  uintptr_t arena_bits = 0;
+  for (size_t offset : offsets) {
+    offset /= sizeof(Arena*);
+    if (offset >= sizeof(arena_bits) * 8) {
+      return std::nullopt;
+    }
+    arena_bits |= uintptr_t{1} << offset;
+  }
+  return arena_bits;
+}
+
 }  // namespace internal
 }  // namespace protobuf
 }  // namespace google
